/*
 * Copyright 2010 Nathanaël Cottin - http://www.insane-network.org
 *
 * This file is part of Insane (Instrumented artificial neural environment).
 *
 * Insane is free software: you can redistribute it and/or modify it under the
 * terms of the Lesser GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or any later version.
 *
 * Insane is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Insane. If not, see <http://www.gnu.org/licenses/>.
 */
package insane.io;

import insane.NetworkLayer;
import insane.NetworkLayerProperties;
import insane.NeuralNetwork;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.net.URL;
import java.util.List;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.xml.sax.SAXException;




/**
 * Defines an XML neural network loader
 * 
 * @author ncottin
 */
public final class XMLNeuralNetworkLoader extends NeuralNetworkLoader {

    private static final String XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";

    @Override
    public NeuralNetwork load(final BufferedReader reader) throws IOException, NeuralNetworkInvalidFormatException {
        Element root;
        try {
            root = readDocument(reader);
        } catch(JDOMException e) {
            throw new NeuralNetworkInvalidFormatException();
        }

        return decode(root);
    }

    /**
     * Creates a neural network based on an XML element
     *
     * @param root The non-<tt>null</tt> element to decode the network from
     * @return The corresponding neural network
     */
    public static NeuralNetwork decode(final Element root) {
        int inputs = Integer.parseInt(root.getAttributeValue("inputs"));
        List<Element> children = root.getChildren();
        NetworkLayerProperties[] properties = new NetworkLayerProperties[children.size()];
        int i = 0;
        int neurons;
        String activation;
        double bias;
        List<Element> subChildren;
        for (Element c : children) {
            subChildren = c.getChildren();
            neurons = subChildren.size();
            activation = c.getAttributeValue("activation");
            bias = Double.parseDouble(c.getAttributeValue("bias"));
            properties[i] = new NetworkLayerProperties(neurons, NeuralNetworkLoader.getActivationFunction(activation), bias);
            ++i;
        }

        NeuralNetwork network = new NeuralNetwork(inputs, properties);
        NetworkLayer[] layers = network.getLayers();
        i = 0;
        int j;
        NetworkLayer layer;
        double[][] weights;
        double[] biases;
        for (Element c : children) {
            layer = layers[i];
            weights = layer.getWeights();
            biases = layer.getBiases();
            subChildren = c.getChildren();
            j = 0;
            for (Element sub : subChildren) {
                biases[j] = Double.parseDouble(sub.getAttributeValue("bias"));
                decodeWeights(sub, weights[j]);
                ++j;
            }

            ++i;
        }

        return network;
    }

    private static void decodeWeights(Element elt, double[] values) {
        List<Element> vals = elt.getChildren();
        int i = 0;
        for (Element val : vals) {
            values[i] = Double.parseDouble(val.getText());
            ++i;
        }
    }

    private static Element readDocument(final BufferedReader reader) throws JDOMException, IOException {
        return new SAXBuilder().build(reader).getRootElement();
    }

    /**
     * Allows to validate the network XML description using an XML schema
     *
     * @param reader The source content to to validated
     * @param xmlSchema The URL of the corresponding schema
     * @throws IOException Either the source content or the XML schema could
     * not be read from their respective sources
     * return <tt>true</tt> only if the given content is valid against the
     * XML schema
     */
    public static boolean isValid(final Reader reader, final URL xmlSchema) throws IOException{
        try {
            SchemaFactory factory = SchemaFactory.newInstance(XML_SCHEMA);
            Schema schema = factory.newSchema(xmlSchema);
            Validator validator = schema.newValidator();
            Source source = new StreamSource(reader);
            validator.validate(source);
        } catch(SAXException e) {
            return false;
        }

        return true;
    }
}